'''Viper Host Groups Demo
~~~~~~~~~~~~~~~~~~~~~~~~~
A host group is a function that produces an instance of :py:class:`viper.collections.Hosts`.


.. tip:: See :py:class:`viper.collections.Host` for more details.


Viper Host Group Definition Structure
-------------------------------------
Here's an example of a viper host group definition.

.. code-block:: python

    def group1() -> Hosts:
        """Just a group of hosts."""

        return Hosts.from_items(
            Host("1.2.3.4"),
        )
'''

from viper import Host
from viper import Hosts


def group1() -> Hosts:
    """Just a group of hosts."""

    return Hosts.from_items(
        Host("127.0.0.1"),
        Host("1.1.1.1"),
        Host("2.2.2.2"),
        Host("8.8.8.8"),
        Host(
            ip="::1",
            hostname="localhost",
            domain="localdomain",
            port=22,
            login_name="root",
            identity_file="/root/.ssh/id_rsa.pub",
        ),
    )



"""Viper Tasks Demo
~~~~~~~~~~~~~~~~~~~
A :py:class:`viper.Task` contains the necessary things that :py:class:`viper.Runner` uses to
generate the command to execute, call the `pre run` and `post run` callbacks,
format the `stdout` and `stderr`, and also time out and retry the execution.

.. tip:: See :py:class:`viper.collections.Task` for more details.
"""

from viper import Task
from viper.demo.callbacks import log_command
from viper.demo.callbacks import log_status
from viper.demo.commands import df_command
from viper.demo.commands import ping_command
from viper.demo.commands import remote_execute_command
from viper.demo.processors import text_stripper


def ping() -> Task:
    """Ping the host one time.

    :rtype: viper.collections.Task
    :example:

    Run the task

    .. code-block:: bash

        viper hosts viper.demo.hosts.group1 \\
                | viper hosts:run-task viper.demo.tasks.ping --max-workers 50 -i 2

    Output

    ::

        8.8.8.8: ('ping', '-c', '1', '8.8.8.8')
        8.8.8.8: OK
        127.0.0.1: ('ping', '-c', '1', '127.0.0.1')
        127.0.0.1: FAILED
        127.0.0.1: ('ping', '-c', '1', '127.0.0.1')
        ...
    """

    return Task(
        "Ping",
        ping_command,
        stdout_processor=text_stripper,
        stderr_processor=text_stripper,
        timeout=5,
        retry=1,
        pre_run=log_command,
        post_run=log_status,
    )


def remote_execute() -> Task:
    """Execute a command on the host.

    :rtype: viper.collections.Task
    :example:

    Run the task

    .. code-block:: bash

        viper hosts viper.demo.hosts.group1 \\
                | viper hosts:run-task viper.demo.tasks.remote_execute "df -h" --max-workers 50 -i 2
    """

    return Task(
        "Remote execute",
        remote_execute_command,
        stdout_processor=text_stripper,
        stderr_processor=text_stripper,
        timeout=300,
        retry=0,
        pre_run=log_command,
        post_run=log_status,
    )


def disk_usage() -> Task:
    """Get the local disk usage of the host

    :rtype: viper.collections.Task
    :example:

    Run the task

    .. code-block:: bash

        viper hosts viper.demo.hosts.group1 \\
                | viper hosts:run-task viper.demo.tasks.disk_usage --max-workers 50 -i 2
    """

    return Task(
        name="Disk Usage",
        command_factory=df_command,
        stdout_processor=text_stripper,
        stderr_processor=text_stripper,
        timeout=5,
        retry=1,
        pre_run=log_command,
        post_run=log_status,
    )



"""Viper Demo Host loaders
~~~~~~~~~~~~~~~~~~~~~~~~~~
These loaders are used with :py:meth:`viper.collections.Hosts.from_file`
to generate a `viper.collections.Hosts` object.


Viper Host Loader Definition Structure
--------------------------------------

.. code-block:: python

    def loader_name(f: typing.TextIO) -> Hosts:

        data = somefunction(f)

        return Hosts.from_list(data)

See :py:func:`viper.demo.loaders.json` for example.
"""

from viper import Hosts

import typing as t


def json(f: t.TextIO) -> Hosts:

    return Hosts.from_json(f.read())



"""Viper Handlers Demo
~~~~~~~~~~~~~~~~~~~~~~
This module contains demo handlers for :py:meth:`viper.collections.pipe`.

Each handler is a function that expects a :py:class:`viper.collections.Collection`
instance as it's first argument.

You then define what to to with that object in that function body.

It's a best practice to return something that can be piped to other commands.
"""

from viper import Hosts
from viper import Results


def hosts_to_csv(hosts: Hosts) -> None:
    """Print the hosts in CSV format

    :param viper.collections.Hosts hosts: The groups of hosts.

    This handler expects a :py:class:`viper.collections.Hosts`
    object and prints the hosts in it in CSV format.

    .. note::
        This handler returns `None`. Thus we won't get any output
        that we can pipe to some other handler or command. This
        is just a demo. Best practice is to always return some
        instance of :py:class:`viper.collections.Collection`.
    """

    print("ip,hostname,domain,port,login_name,identity_file")
    for host in hosts.all():
        print(
            host.ip,
            host.hostname or "",
            host.domain or "",
            host.port or "",
            host.login_name or "",
            host.identity_file or "",
            sep=",",
        )


def print_status(results: Results) -> None:
    """Print the status of the results in the terminal.

    :param viper.collections.Results results: The groups of result.

    This handler expects a :py:class:`viper.collections.Result`
    object and prints the results in it in CSV format.

    .. note::
        This handler also returns `None`. Thus we won't get any output
        that we can pipe to some other handler or command. This
        is just a demo. Best practice is to always return some
        instance of :py:class:`viper.collections.Collection`.
    """

    for result in results.all():
        if result.errored():
            print(f"{result.task.name}: {result.host.ip} FAILED")
        else:
            print(f"{result.task.name}: {result.host.ip} PASSED")


def export_csv(results: Results, csv_file: str) -> Results:
    """Export the result to a CSV file location.

    :param viper.collections.Results results: The groups of result.

    This handler expects a :py:class:`viper.collections.Result`
    object and prints the results in it in CSV format.

    .. note::
        Unlike :py:class:`viper.demo.handlers.print_status`, this
        handlers returns an instance of `viper.collections.Results`
        object with can be piped to further commands and handlers.

        This is the recommended way to defile handlers.
    """

    import csv
    from time import ctime

    with open(csv_file, "w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(
            [
                "task",
                "ip",
                "hostname",
                "command",
                "returncode",
                "success",
                "start",
                "end",
                "retry",
                "stdout",
                "stderr",
            ]
        )
        for r in results.all():
            writer.writerow(
                [
                    r.task.name,
                    r.host.ip,
                    r.host.hostname,
                    r.command,
                    r.returncode,
                    r.ok(),
                    ctime(r.start),
                    ctime(r.end),
                    r.retry,
                    r.stdout,
                    r.stderr,
                ]
            )

    # Let's return the result to stdout so that it can be piped to other commands
    return results



"""Viper Jobs Demo
~~~~~~~~~~~~~~~~~~
A viper job is a custom defined workflow. Unlike task, a job may contain
multiple tasks connected together with handlers, filters etc.

A job recieves a :py:class:`viper.collections.Hosts` object (read from stdin)
and should always return a :py:class:`viper.collections.Results` object. You can define arguments
which will be passed as `str` when you run it. You define what to do with the args
in the function body and then return the result.


Viper Job Definition Structure
------------------------------

.. code-block:: python

    def job_name(hosts, *args: str) -> Results:
        # Do something and return Results

You can run it with

.. code-block:: bash

    cat hosts.json | viper run job_name
"""

from viper import Hosts
from viper import Results
from viper.demo.handlers import export_csv
from viper.demo.tasks import ping
from viper.demo.tasks import remote_execute


def ping_then_execute(hosts: Hosts, command: str, resultsfile: str) -> Results:
    """First ping the hosts, then run the given command of pingable hosts.

    :param str data: Input data as `str`.
    :param str command: Command to execute.
    :param str resultsfile: Results will be saved in this file in CSV format.

    :rtype: viper.Results

    :example:

    .. code-block:: bash

        viper hosts viper.demo.hosts.group1 \\
                | viper run-job viper.demo.jobs.ping_then_execute "df -h" results.csv
    """
    return (
        hosts.run_task(ping(), max_workers=50)
        .filter(lambda result: result.ok())
        .hosts()
        .run_task(remote_execute(), command, max_workers=50)
        .pipe(export_csv, resultsfile)
    )



from __future__ import annotations
from dataclasses import dataclass
from dataclasses import field
from sqlite3 import Cursor
from viper.const import Config

import sqlite3
import typing as t


@dataclass
class ViperDB:

    url: str = Config.db_url.value
    engine: t.Optional[sqlite3.Connection] = field(init=False, default=None)

    @classmethod
    def init(cls, url: str, force: bool = False) -> None:
        if force:
            with cls(url) as conn:
                conn.execute("DROP TABLE IF EXISTS results")

        with cls(url) as conn:
            conn.execute(
                """
                CREATE TABLE results (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    hash INTEGER NOT NULL,
                    trigger_time REAL NOT NULL,
                    task JSON NOT NULL,
                    host JSON NOT NULL,
                    args JSON NOT NULL,
                    command JSON NOT NULL,
                    stdout BLOB NOT NULL,
                    stderr BLOB NOT NULL,
                    returncode INTEGER NOT NULL,
                    start REAL NOT NULL,
                    end REAL NOT NULL,
                    retry INTEGER NOT NULL
                );
                """
            )

    def __enter__(self) -> Cursor:
        self.engine = sqlite3.connect(self.url)
        return self.engine.cursor()

    def __exit__(self, exc_type: type, exc_value: t.Any, exc_traceback: Exception):
        try:
            self.engine.commit()
        except Exception:  # pragma: no cover
            self.engine.rollback()
        self.engine.close()



"""Base classes for the viper.cli module."""

from argparse import _SubParsersAction
from argparse import ArgumentParser
from argparse import Namespace

import typing as t

__all__ = ["SubCommand"]


class SubCommand:
    """Base class for the subcommand parsers."""

    name: t.Optional[str] = None

    aliases: t.Sequence[str] = ()

    @classmethod
    def attach_to(cls, subparsers: _SubParsersAction) -> None:

        if not cls.name:
            raise NotImplementedError()  # no cover

        subparser = subparsers.add_parser(cls.name, help=cls.__doc__)
        if cls.aliases:
            for alias in cls.aliases:
                cls(subparsers.add_parser(alias, help=f"alias of {repr(cls.name)}"))
        cls(subparser)

    def __init__(self, subparser: ArgumentParser) -> None:
        self.add_arguments(subparser)
        subparser.set_defaults(_handler=self)
        subparser.add_argument(
            "--debug",
            action="store_true",
            help="show traceback information when an exception is raised",
        )

    def add_arguments(self, parser: ArgumentParser) -> None:
        raise NotImplementedError()  # no cover

    def __call__(self, args: Namespace) -> int:
        raise NotImplementedError()  # no cover



"""Viperfile Demo
~~~~~~~~~~~~~~~~~
A ``viperfile.py`` let's you extend the Viper CLI commands and options.
The host groups, filters, handlers, jobs etc. you define in the file,
will be added to the Viper CLI's command list for easy access and documentation.

The ``viperfile.py`` file should be in your current working directory.


Example Viperfile CLI Usage
---------------------------

.. code-block:: bash

    viper --help

    viper @myproj:allhosts --help

    viper @myproj:allhosts | \\
            viper @myproj:hosts_by ip 127.0.0.1 | \\
            viper @myproj:remote_exec "df -h" results.csv --max-workers 50

.. tip:: See :py:mod:`viper.project` for more details on available project APIs.
"""

from argparse import FileType
from argparse import Namespace
from time import ctime
from viper import Host
from viper import Hosts
from viper import Result
from viper import Results
from viper import Runner
from viper import Task
from viper.project import arg
from viper.project import Project

import csv
import json
import sys
import typing as t

myproj = Project(prefix="myproj")


def log_command_callback(runner: Runner) -> None:
    """Log the command before run.

    :param viper.collections.Runner runner: Pre-run callbacks are given this object.
    :rtype: None
    """

    command = runner.task.command_factory(runner.host, *runner.args)
    print(runner.host.ip, command, file=sys.stderr, sep=": ")


def log_status_callback(result: Result) -> None:
    """Log the status after task run.

    :param viper.collections.Result result: Post-run callbacks are given this object.
    :rtype: None
    """

    if result.ok():
        print(result.host.ip, "OK", file=sys.stderr, sep=": ")
        return

    print(result.host.ip, "FAILED", file=sys.stderr, sep=": ")


def text_stripper(txt: str) -> str:
    """This is a stdout/stderr processor that strips the given text.

    :param str txt: The text to be stripped
    :rtype: str
    """
    return txt.strip()


@myproj.hostgroup(
    args=[
        arg("-f", "--file", type=FileType("r"), default="hosts.json"),
        arg("-I", "--identity_file", default="/root/.ssh/id_rsa.pub"),
    ]
)
def allhosts(args: Namespace) -> Hosts:
    """Get all the myproj hosts

    :param Namespace args: The parsed arguments.
    :rtype: viper.collections.Hosts
    :example:

    .. code-block:: bash

        viper @myproj:allhosts
    """

    data = json.load(args.file)

    return Hosts.from_items(
        Host(
            ip=d["ip"],
            hostname=d["name"],
            login_name="root",
            identity_file=args.identity_file,
            meta=tuple(d.items()),
        )
        for d in data
    )


@myproj.filter(objtype=Hosts, args=[arg("key"), arg("val")])
def hosts_by(host: Host, args: Namespace) -> bool:
    """Filter hosts by key and metadata

    :param viper.collections.Host host: Host filters are given this object.
    :param Namespace args: The parsed arguments.
    :rtype: bool
    :example:

    .. code-block:: bash

        viper @myproj:allhosts \\
                | viper @myproj:hosts_by ip 1.1.1.1

    .. note: We can also use the `hosts:where` command.
    """

    return str(dict(host.meta)[args.key]) == args.val


@myproj.filter(objtype=Results, args=[arg("key"), arg("val")])
def results_by(result: Result, args: Namespace) -> bool:
    """Filter hosts by IP address

    :param viper.collections.Result result: Result filters are given this object.
    :param Namespace args: The parsed arguments.
    :rtype: bool
    :example:

    .. code-block:: bash

        viper results \\
                | viper @myproj:results_by returncode 0

    .. note: We can also use the `results:where` command.
    """

    return (
        hasattr(result, args.key) and str(getattr(result, args.key)).strip() == args.val
    )


@myproj.handler(fromtype=Hosts, totype=Hosts, args=[arg("file", type=FileType("w"))])
def hosts2csv(hosts: Hosts, args: Namespace) -> Hosts:
    """Export the hosts to a CSV file

    :param viper.collections.Hosts hosts: This will be read from stdin.
    :param Namespace args: The parsed arguments.
    :rtype: viper.collections.Hosts
    :example:

    .. code-block:: bash

        viper @myproj:allhosts \\
                | viper @myproj:hosts2csv hosts.csv
    """

    writer = csv.writer(args.file)
    writer.writerow(list(dict(hosts.all()[0].meta).keys()))
    for host in hosts.all():
        writer.writerow(list(dict(host.meta).values()))
    args.file.flush()
    args.file.close()

    return hosts


@myproj.handler(
    fromtype=Results, totype=Results, args=[arg("file", type=FileType("w"))]
)
def results2csv(results: Results, args: Namespace) -> Results:
    """Export the results to a CSV file

    :param viper.collections.Results results: This will be read from stdin.
    :param Namespace args: The parsed arguments.
    :rtype: viper.collections.Results
    :example:

    .. code-block:: bash

        viper results \\
                | viper @myproj:results2csv results.csv
    """

    writer = csv.writer(args.file)
    writer.writerow(
        [
            "task",
            "ip",
            "hostname",
            "command",
            "returncode",
            "success",
            "start",
            "end",
            "retry",
            "stdout",
            "stderr",
        ]
    )

    for r in results.all():
        writer.writerow(
            [
                r.task.name,
                r.host.ip,
                r.host.hostname,
                r.command,
                r.returncode,
                r.ok(),
                ctime(r.start),
                ctime(r.end),
                r.retry,
                r.stdout,
                r.stderr,
            ]
        )
    args.file.flush()
    args.file.close()

    # Printing JSON to terminal to that it can be piped to further commands
    return results


def remote_exec_command(host: Host, command: str) -> t.Sequence[str]:
    """Command for the `remote_exec` job

    :param viper.collections.Host host: This will be passed by the task.
    :param str command: The command to execute.
    :rtype: tuple
    """
    if not host.login_name:
        raise ValueError(f"{host}: 'login_name' is not set")

    if not host.identity_file:
        raise ValueError(f"{host}: 'identity_file' is not set")

    return (
        "ssh",
        "-l",
        host.login_name,
        "-i",
        host.identity_file,
        "-o",
        "StrictHostKeyChecking=no",
        host.ip,
        command,
    )


@myproj.job(
    args=[
        arg("command"),
        arg("file", type=FileType("w"), help="CSV file path for the result"),
        arg("--max-workers", default=0, type=int),
    ],
)
def remote_exec(hosts: Hosts, args: Namespace) -> Results:
    """Execute command on a remote host via SSH

    :param viper.collections.Host host: This will be read from stdin.
    :param Namespace args: The parsed arguments.
    :rtype: viper.collections.Results
    :example:

    .. code-block:: bash

        viper @myproj:allhosts \\
                | viper @myproj:remote_exec "df -h" results.csv --max-workers 50
    """
    return (
        hosts.task(
            Task(
                "Remote execute",
                remote_exec_command,
                timeout=300,
                retry=0,
                pre_run=log_command_callback,
                post_run=log_status_callback,
            ),
            args.command,
        )
        .run(max_workers=args.max_workers)
        .final()  # Filter only the final results
        .pipe(lambda results: results2csv(results, args))
    )


def app_version_command(host: Host, app: str) -> t.Sequence[str]:
    """Command for the `app_version` job

    :param viper.collections.Host host: This will be given by the task.
    :param str app: The application binary name.
    :rtype: tuple
    """
    return remote_exec_command(host, f"{app} --version")


@myproj.job(
    args=[
        arg("app"),
        arg("file", type=FileType("w"), help="CSV file path for the result"),
        arg("--max-workers", default=0, type=int),
    ],
)
def app_version(hosts: Hosts, args: Namespace) -> Results:
    """Gets the version of an application installed on a remote host

    :param viper.collections.Host: Jobs are given hosts read from stdin.
    :param Namespace args: The parsed arguments.
    :rtype: viper.collections.Results
    :example:

    .. code-block:: bash

        viper @myproj:allhosts \\
                | viper @myproj:app_version python results.csv --max-workers 50
    """
    return (
        hosts.task(
            Task(
                "Get app version",
                app_version_command,
                timeout=10,
                retry=3,
                pre_run=log_command_callback,
                post_run=log_status_callback,
            ),
            args.app,
        )
        .run(max_workers=args.max_workers)
        .final()  # Filter only the final results
        .pipe(lambda results: results2csv(results, args))
    )


def install_via_apt_command(host: Host, app: str, version: str) -> t.Sequence[str]:
    """Generates command for the `install_via_apt` job

    :param viper.collections.Host host: This will be given by the task.
    :param str app: The target application name.
    :param str app: The target application version.
    :rtype: tuple
    """
    return remote_exec_command(host, f"apt-get install -f {app}-{version}")


@myproj.job(
    args=[
        arg("app"),
        arg("version"),
        arg("file", type=FileType("w"), help="CSV file path for the result"),
        arg("--max-workers", default=0, type=int),
    ],
)
def install_via_apt(hosts: Hosts, args: Namespace) -> Results:
    """First checks if the specific version of app is installed, if not, installs it

    :param viper.collections.Host host: This will be given by the task.
    :param Namespace args: The parsed arguments.
    :rtype: viper.collections.Results
    :example:

    .. code-block::

        viper @myproj:allhosts \\
                | viper @myproj:install_via_apt python 3.8.0 results.csv --max-workers 50
    """
    return (
        hosts.task(
            Task(
                "Get app version",
                app_version_command,
                timeout=10,
                retry=3,
                pre_run=log_command_callback,
                post_run=log_status_callback,
            ),
            args.app,
        )
        .run(max_workers=args.max_workers)
        .final()  # Filter only the final results
        .filter(lambda results: results.ok() and args.version not in results.stdout)
        .hosts()
        .task(
            Task(
                "Install app with specific version",
                install_via_apt_command,
                timeout=300,
                retry=5,
                pre_run=log_command_callback,
                post_run=log_status_callback,
            ),
            args.app,
            args.version,
        )
        .run(max_workers=args.max_workers)
        .final()  # Filter only the final results
        .pipe(lambda results: results2csv(results, args))
    )


@myproj.action()
def get_triggers(args: Namespace) -> t.Sequence[float]:
    """Get the unique trigger times from history"""

    results = Results.from_history(final=True).all()
    triggers = set(x.trigger_time for x in results)
    return tuple(sorted(triggers))



"""Viper Demo Output Processors

This module contains example of output processors to be used in task definition.


.. tip:: See :py:class:`viper.collections.Task` and :py:class:`viper.demo.tasks.ping`.
"""


def text_stripper(txt: str) -> str:
    """This is a stdout/stderr processor that strips the given text.

    :param str txt: The text to be stripped
    :rtype: str
    """
    return txt.strip()



"""Viper Actions Demo
~~~~~~~~~~~~~~~~~~~~~
A viper action is a plain Python function that viper executes when called.
An action takes optional arguments passed while invoking the command and can return
anything. If it returns ``None``, nothing will be printed to stdout. Otherwise the
string representation of the returned object will be printed.

If multiple objets are returned, they will be printed in separate lines.

Viper Action Definition Structure
---------------------------------

.. code-block:: python

    def action_name(*args: str) -> object:
        # Do something and return anything

You can run it with

.. code-block:: bash

    viper lets action_name
"""

from viper import Results

import typing as t


def get_triggers() -> t.Sequence[float]:
    """Get the unique trigger times from history"""

    results = Results.from_history(final=True).all()
    triggers = set(x.trigger_time for x in results)
    return tuple(sorted(triggers))



__author__ = "Arijit Basu"
__description__ = "Viper is a handy tool for easily running infrastructure management tasks and commands."
__email__ = "sayanarijit@gmail.com"
__homepage__ = "https://github.com/sayanarijit/viper"
__license__ = "MIT"
__version__ = "v0.26.1"

from viper.collections import Host  # noqa: F401
from viper.collections import Hosts  # noqa: F401
from viper.collections import Result  # noqa: F401
from viper.collections import Results  # noqa: F401
from viper.collections import Runner  # noqa: F401
from viper.collections import Runners  # noqa: F401
from viper.collections import Task  # noqa: F401
from viper.collections import WhereConditions  # noqa: F401

__all__ = [
    "Host",
    "Hosts",
    "Result",
    "Results",
    "Runner",
    "Runners",
    "Task",
    "WhereConditions",
]
__doc__ = f"""::

     ▄   ▄█ █ ▄▄  ▄███▄   █▄▄▄▄   ▄█    ▄   ▄████  █▄▄▄▄ ██     ▄█▄    ████▄ █▀▄▀█ █▀▄▀█ ██      ▄   ██▄   ▄███▄   █▄▄▄▄
      █  ██ █   █ █▀   ▀  █  ▄▀   ██     █  █▀   ▀ █  ▄▀ █ █    █▀ ▀▄  █   █ █ █ █ █ █ █ █ █      █  █  █  █▀   ▀  █  ▄▀
 █     █ ██ █▀▀▀  ██▄▄    █▀▀▌    ██ ██   █ █▀▀    █▀▀▌  █▄▄█   █   ▀  █   █ █ ▄ █ █ ▄ █ █▄▄█ ██   █ █   █ ██▄▄    █▀▀▌
  █    █ ▐█ █     █▄   ▄▀ █  █    ▐█ █ █  █ █      █  █  █  █   █▄  ▄▀ ▀████ █   █ █   █ █  █ █ █  █ █  █  █▄   ▄▀ █  █
   █  █   ▐  █    ▀███▀     █      ▐ █  █ █  █       █      █   ▀███▀           █     █     █ █  █ █ ███▀  ▀███▀     █
    █▐        ▀            ▀         █   ██   ▀     ▀      █                   ▀     ▀     █  █   ██                ▀
    ▐                                                     ▀                               ▀


.. image:: https://img.shields.io/pypi/v/viper-infra-commander.svg
    :target: https://pypi.org/project/viper-infra-commander

.. image:: https://img.shields.io/pypi/pyversions/viper-infra-commander.svg
    :target: https://pypi.org/project/viper-infra-commander

.. image:: https://travis-ci.com/sayanarijit/viper.svg?branch=master
    :target: https://travis-ci.com/sayanarijit/viper

.. image:: https://codecov.io/gh/sayanarijit/viper/branch/master/graph/badge.svg
    :target: https://codecov.io/gh/sayanarijit/viper

.. image:: https://img.shields.io/badge/code%20style-black-000000.svg
    :target: https://github.com/python/black

.. image:: https://readthedocs.org/projects/viper-infrastructure-commander/badge/?version=latest
    :target: https://viper-infrastructure-commander.readthedocs.io


{__description__}


Installation
~~~~~~~~~~~~

.. code-block:: bash

    pip install -U viper-infra-commander

    # Or with tab autocomplete support

    pip install -U "viper-infra-commander[autocomplete]"


Getting Started
~~~~~~~~~~~~~~~

.. code-block:: bash

    # See the help menu
    viper -h

    # Initialize SQLite DB
    viper init -f

    # Run a job on a defined list of hosts
    viper hosts viper.demo.hosts.group1 \\
            | viper run-job viper.demo.jobs.ping_then_execute "df -h" results.csv


Further Readings
~~~~~~~~~~~~~~~~
**API Docs with Examples ☞** https://viper-infrastructure-commander.readthedocs.io


Contributing To Viper
~~~~~~~~~~~~~~~~~~~~~
**Contribution Guidelines ☞** https://github.com/sayanarijit/viper/blob/master/CONTRIBUTING.md
"""



"""Datatypes for Viper Objects
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
The data types for Viper objects are defined here.
"""

from __future__ import annotations
from collections import OrderedDict
from collections.abc import Iterable
from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from json import dumps as dumpjson
from json import loads as loadjson
from pydoc import locate
from time import time
from viper.const import Config
from viper.db import ViperDB

import subprocess
import sys
import traceback
import typing as t

CollectionType = t.TypeVar("CollectionType", bound="Collection")
ItemType = t.TypeVar("ItemType", bound="Item")
ItemsType = t.TypeVar("ItemsType", bound="Items")

__all__ = [
    "WhereConditions",
    "Collection",
    "Item",
    "Items",
    "Host",
    "Hosts",
    "Task",
    "Runner",
    "Runners",
    "Result",
    "Results",
]


class HandlerType:
    """TODO: This should be a protocol"""

    def __call__(self, obj: object, *args: object) -> object:  # pragma: no cover
        pass


class FilterType:
    """TODO: This should be a protocol"""

    def __call__(self, obj: object, *args: object) -> bool:  # pragma: no cover
        pass


class CommandFactoryType:
    """TODO: This should be a protocol"""

    pass


class WhereConditions(Enum):
    """Where query conditions for viper Items"""

    is_ = "IS"
    is_not = "IS_NOT"
    contains = "CONTAINS"
    not_contains = "NOT_CONTAINS"
    startswith = "STARTSWITH"
    not_startswith = "NOT_STARTSWITH"
    endswith = "ENDSWITH"
    not_endswith = "NOT_ENDSWITH"


@dataclass(frozen=True, order=True)
class Collection:
    """The base collection class.

    This is the parent class for all Viper native objects such as
    :py:class:`viper.collections.Host`, :py:class:`viper.collections.Results` etc.
    """

    def __str__(self):
        return self.to_json()

    def from_json(
        cls: t.Type[CollectionType], json: str, *args: object, **kwargs: object
    ) -> CollectionType:  # pragma: no cover
        """Initialise a new object of this class from the given JSON string.

        :param str json: The JSON data to parse.
        :param object args and kwargs: These will be passed to `json.laods`.

        :rtype: viper.collections.Collection

        :example:

        .. code-block:: python

            Hosts.from_json('[{"ip": "1.1.1.1"}]')
        """
        raise NotImplementedError()

    def to_json(self, *args: object, **kwargs: object) -> str:  # pragma: no cover
        """Represent the collection as JSON data.

        :param object args and kwargs: These will be passed to `json.laods`.

        :rtype: str

        :example:

        .. code-block:: python

            Host("1.2.3.4").to_json(indent=4)
        """
        raise NotImplementedError()

    @classmethod
    def from_func(cls: t.Type[CollectionType], funcpath: str) -> CollectionType:
        """Load the object from the given Python function.

        :param str funcpath: The path to a Python function that returns an
            instance of this class.

        :rtype: Collection

        :example:

        .. code-block:: python

            Task.from_func(ping)

        .. tip:: See :py:class:`viper.demo.tasks.ping`
        """

        func: t.Optional[t.Callable[[], CollectionType]] = locate(funcpath)

        if not func:
            raise ValueError(f"could not resolve {repr(funcpath)}.")

        obj = func()
        if not isinstance(obj, cls):
            raise ValueError(
                f"{repr(funcpath)} does not produce a valid {cls} instance."
            )

        return obj

    def pipe(self: CollectionType, handler: HandlerType, *args: object) -> object:
        """Pipe this object to the given function.

        :param callable handler: A callable that expects this object.
        :param object args: Arguments to be passed to the handler for decision making.

        :example:

        .. code-block:: python

            Hosts.from_items(Host("1.2.3.4")).pipe(hosts2csv)

        .. tip:: See :py:func:`viper.demo.handlers.hosts2csv`
        """
        return handler(self, *args)

    def hash(self: CollectionType) -> int:
        """Get the hash value.

        By convention all Viper native objects are frozen and thus hashable.

        :rtype: int
        """
        return hash(self)


@dataclass(frozen=True, order=True)
class Item(Collection):
    """The base class for a single Viper object.

    :py:class:`viper.collections.Host`, :py:class:`viper.collections.Task` etc. inherits this base class.
    """

    @classmethod
    def from_dict(  # pragma: no cover
        cls: t.Type[ItemType], dict_: t.Dict[str, object]
    ) -> ItemType:
        """Initialize item from the given dict.

        :param dict dict_: The dictionary containing the properties for this object.

        :rtype: viper.collections.Item

        :example:

        .. code-block:: python

            Host.from_dict({"ip": "1.2.3.4"})
        """
        raise NotImplementedError()

    def to_dict(self: ItemType) -> t.Dict[str, object]:  # pragma: no cover
        """Represent the item as dict.

        :rtype: dict

        :example:

        .. code-block:: python

            Host("1.2.3.4").to_dict()
        """
        raise NotImplementedError()

    @classmethod
    def from_json(cls: t.Type[ItemType], json, *args, **kwargs) -> ItemType:
        return cls.from_dict(loadjson(json, *args, **kwargs))

    def to_json(self: ItemType, *args, **kwargs) -> str:
        return dumpjson(self.to_dict(), *args, **kwargs)

    def format(self: Item, template: str) -> str:
        """Get a custom string representation of this object.

        :param str template: The template to be used as `"template".format(**self.to_dict())`
        :rtype: str

        :example:

        .. code-block:: python

            Host("1.2.3.4").format("{ip} {hostname} {meta[tag]}")
        """
        return template.format(**self.to_dict())


@dataclass(frozen=True)
class Items(Collection):
    """The base class for collection objects for a group of similar items.

    :py:class:`viper.collections.Hosts`, :py:class:`viper.collections.Results` etc. inherits this base class.
    """

    _all: t.Sequence[Item] = ()
    _item_factory: t.Type[Item] = field(init=False)

    @classmethod
    def from_items(
        cls: t.Type[Item], *items: t.Union[t.Iterable[Item], Item]
    ) -> ItemsType:
        """Create this an instance of this object using the given items.

        :param viper.collections.Item items: The group of items to hold.

        :rtype: viper.collections.Items

        .. note::
            Classes that inherits from :py:class:`Items` should not be
            initialized directly (e.g. ``Hosts(Host("1.2.3.4"))``). Factory
            methods such as this should be used instead.

        :example:

        .. code:: python

            Hosts.from_items(Host("1.2.3.4"))
        """
        items_set = OrderedDict()
        for item in items:
            if not isinstance(item, Item) and not isinstance(item, Iterable):
                raise ValueError(
                    f"{item}: expecting 'Item' or generator of 'Item's but got {type(item)}"
                )

            if isinstance(item, Item):
                items_set[item] = None
                continue

            if isinstance(item, Iterable):
                for i in item:
                    if not isinstance(i, Item):
                        raise ValueError(f"{i}: expecting 'Item' but got {type(i)}")
                    items_set[i] = None

        return cls(tuple(items_set))

    @classmethod
    def from_list(
        cls: t.Type[ItemsType], list_: t.Sequence[t.Dict[str, object]]
    ) -> ItemsType:
        """Initialize items from given list of dictiionaries.

        :param list list_: The list of dictiionaries containing the item properties.

        :rtype: viper.collections.Items

        This is used for loading JSON data into an instance of this class.

        :example:

        .. code:: python

            Hosts.from_list([{"ip": "1.2.3.4"}])
        """

        if cls._item_factory is None:  # pragma: no cover
            raise NotImplementedError()

        try:
            return cls.from_items(map(cls._item_factory.from_dict, list_))
        except Exception:
            raise ValueError(f"invalid input data for {cls.__name__}")

    def to_list(self: ItemsType) -> t.List[t.Dict[str, object]]:
        """Represent the items as a list of dictiionaries.

        This is used for dumping an instance of this object as JSON data.

        :rtype: list

        :example:

        .. code:: python

            Hosts.from_items(Host("1.2.3.4")).to_list()
        """

        return [i.to_dict() for i in self._all]

    @classmethod
    def from_json(
        cls: t.Type[ItemsType], json: str, *args: object, **kwargs: object
    ) -> ItemsType:

        return cls.from_list(loadjson(json))

    def to_json(self: ItemsType, *args: object, **kwargs: object) -> str:

        return dumpjson(self.to_list(), *args, **kwargs)

    def __len__(self: ItemsType) -> int:
        return len(self._all)

    def count(self: ItemsType) -> int:
        """Count the number of items.

        :rtype: int
        """
        return len(self)

    def head(self: ItemsType, n: int = 10) -> ItemsType:
        """Get the first 'n' items from the group of items.

        Works like Python's [:n] index

        :rtype: viper.collections.Item
        """
        return self.range(None, n)

    def tail(self: ItemsType, n: int = 10) -> ItemsType:
        """Get the last 'n' item from the group of items.

        Works like Python's [-n:] index

        :rtype: viper.collections.Item
        """
        return self.range(-n)

    def range(
        self: ItemsType, i: t.Optional[int] = None, j: t.Optional[int] = None
    ) -> ItemsType:
        """get the items in given range.

        :param int i (optional): The first index.
        :param int j (optional): The last index.

        :rtype: Items

        This has the same behaviour as Python's [i:j]
        """
        return type(self).from_items(*self._all[i:j])

    def sort(
        self: ItemsType, key: t.Optional[t.Callable[[Item], object]] = None
    ) -> ItemsType:
        """Sort the items by given key/function.

        :param callable key: A function similar to the one passed to the built-in `sorted()`.
        """
        return type(self)(tuple(sorted(self._all, key=key)))

    def filter(self: ItemsType, filter_: FilterType, *args: object) -> ItemsType:
        """Filter the items by a given function.

        :param callable filter_: A function that returns either True or False.
        :param object args: Arguments to be passed to the filter to manipulate the decision making.
        """
        return type(self)(tuple(filter(lambda i: filter_(i, *args), self._all)))

    def all(self: ItemsType) -> t.Sequence[Item]:
        """Get a tuple of all the items.

        :rtype: tuple
        """
        return self._all

    def format(self: ItemsType, template: str, sep="\n") -> str:
        """Get a custom string representation of this list of objects.

        :param str template: The template will be compiled by Python's `.format()`.
        :param str sep: The separator used to separate all the items.

        :rtype: str

        :example:

        .. code-block:: python

            Hosts.from_items(Host("1.2.3.4")).format("{ip} {hostname} {meta[tag]}")
        """
        return sep.join(template.format(**x.to_dict()) for x in self._all)

    def where(
        self, key: str, condition: WhereConditions, values: t.Sequence[str]
    ) -> ItemsType:
        """Select items by a custom query.

        :param str key: The key will be compiled by Python's `.format()`.
        :param viper.collections.WhereConditions condition: The where condition.
        :param list values: The values for the key.

        :rtype: Items
        """
        result = []

        for obj in self._all:
            val = f"{{{key}}}".format(**obj.to_dict())

            if condition is WhereConditions.is_:
                if val in values:
                    result.append(obj)

            elif condition is WhereConditions.is_not:
                if val not in values:
                    result.append(obj)

            elif condition is WhereConditions.contains:
                if any(map(lambda x: x in val, values)):
                    result.append(obj)

            elif condition is WhereConditions.not_contains:
                if any(map(lambda x: x not in val, values)):
                    result.append(obj)

            elif condition is WhereConditions.startswith:
                if any(map(lambda x: val.startswith(x), values)):
                    result.append(obj)

            elif condition is WhereConditions.not_startswith:
                if any(map(lambda x: not val.startswith(x), values)):
                    result.append(obj)

            elif condition is WhereConditions.endswith:
                if any(map(lambda x: val.endswith(x), values)):
                    result.append(obj)

            elif condition is WhereConditions.not_endswith:
                if any(map(lambda x: not val.endswith(x), values)):
                    result.append(obj)

            else:
                raise ValueError(f"expecting enum {WhereConditions}")

        return type(self).from_items(*result)


@dataclass(frozen=True, order=True)
class Host(Item):
    """Viper Host class."""

    ip: str
    hostname: t.Optional[str] = None
    domain: t.Optional[str] = None
    port: int = 22
    login_name: t.Optional[str] = None
    identity_file: t.Optional[str] = None
    meta: t.Sequence[t.Tuple[object, object]] = ()

    def __hash__(self) -> int:
        return hash(self.ip)

    def __eq__(self, value):
        return type(self) == type(value) and self.ip == value.ip

    @classmethod
    def from_dict(cls, dict_):
        try:
            host = cls(**dict(dict_, meta=tuple(dict_.get("meta", {}).items())))
        except Exception:
            raise ValueError(f"invalid input data for {cls.__name__}")
        return host

    def to_dict(self):
        return dict(vars(self), meta=dict(self.meta))

    def fqdn(self) -> str:
        """Get the FQDN from hostname and domain name.

        :rtype: str

        :raises ValueError: If either of hostname or domain name is not set.
        """

        if not self.hostname and not self.domain:
            raise AttributeError("hostname and domain not set")

        if not self.hostname:
            raise AttributeError("hostname not set")

        if not self.domain:
            raise AttributeError("domain not set")

        return f"{self.hostname}.{self.domain}"

    def task(self, task: Task, *args: str) -> Runner:
        """Assigns a task to be run.

        :param viper.collections.Task task: The task to be assigned.
        :param str args: The arguments to be used to create the command from `command_factory`.

        :rtype: viper.collections.Runner

        :example:

        .. code-block:: python

            Host("1.2.3.4").task(ping)

        .. tip:: See :py:class:`viper.demo.tasks.ping`
        """

        return Runner(task=task, host=self, args=args)

    def run_task(self, task: Task, *args: str) -> Result:
        """Assign the task to the host and then run it.

        :param viper.collections.Task task: The task to be assigned.
        :param str args: The arguments to be used to create the command
            from :py:attr:`viper.collections.Task.command_factory`.

        :rtype: viper.collections.Result

        :example:

        .. code-block:: python

            Host("1.2.3.4").task_task(ping)

        .. tip:: See :py:class:`viper.demo.tasks.ping`
        """

        return self.task(task, *args).run()

    def results(self) -> Results:
        """Fetch recent results of current host from database."""
        return self.pipe(Results.by_host)


@dataclass(frozen=True)
class Hosts(Items):
    """A group of :py:class:`viper.collections.Host` objects.

    .. tip:: See :py:mod:`viper.demo.hosts`
    """

    _item_factory: t.Type[Host] = field(init=False, default=Host)

    @classmethod
    def from_file(
        cls, filepath: str, loader: t.Optional[t.Callable[[t.TextIO], "Hosts"]] = None
    ) -> Hosts:
        """Initialize hosts by reading data from a file.

        :param filepath str: The path for the file to read from.
        :param callable loader (optional): A custom loader.

        :rtype: viper.collections.Hosts

        :example:

        .. code-block:: python

            Host("1.2.3.4").from_file("/path/to/file/hosts.json", loader=viper.demo.loader.json)

        .. tip:: See :py:func:`viper.demo.loaders.json`
        """

        if loader is None:

            def _loader(f: t.TextIO) -> Hosts:
                return cls.from_items(
                    Host(ip.strip()) for ip in f.read().strip().split()
                )

            loader = _loader

        with open(filepath) as f:
            return loader(f)

    def task(self, task: Task, *args: str) -> Runners:
        """Assigns a task to be run on each host in the group.

        :param viper.collections.Task task: The task to be assigned.
        :param str args: The arguments to be used to create the command from `command_factory`.

        :rtype: viper.collections.Runners

        :example:

        .. code-block:: python

            Hosts.from_items(Host("1.2.3.4")).task(ping)

        .. tip:: See :py:class:`viper.demo.tasks.ping`
        """

        return Runners.from_items(
            Runner(task=task, host=h, args=args) for h in self._all
        )

    def run_task(
        self, task: Task, *args: str, max_workers=Config.max_workers.value
    ) -> Results:
        """Assign the task to the host and then run it.

        :param viper.collections.Task task: The task to be assigned.
        :param str args: The arguments to be used to create the command
            from :py:attr:`viper.collections.Task.command_factory`.
        :param int max_workers: Maximum number of thread workers.
            if the value is <= 1, tasks will run in sequence.

        :rtype: viper.collections.Results

        :example:

        .. code-block:: python

            Hosts.from_items(Host("1.2.3.4")).task_task(ping)

        .. tip:: See :py:class:`viper.demo.tasks.ping`
        """
        return self.task(task, *args).run(max_workers=max_workers)

    def results(self) -> Results:
        """Get the past results of this group of hosts from history.

        :rtype: viper.collections.Results
        """
        results = []
        for h in self._all:
            for r in h.results().all():
                results.append(r)
        return Results.from_items(*results)


@dataclass(frozen=True, order=True)
class Task(Item):
    """A task is a definition of a single operation.

    A task must contain name and a **named function** to generate the
    command.

    .. tip:: See :py:mod:`viper.demo.tasks`
    """

    name: str
    command_factory: CommandFactoryType
    timeout: t.Optional[int] = None
    retry: int = 0
    stdout_processor: t.Optional[t.Callable[[str], str]] = None
    stderr_processor: t.Optional[t.Callable[[str], str]] = None
    pre_run: t.Optional[t.Callable[[Runner], None]] = None
    post_run: t.Optional[t.Callable[[Result], None]] = None
    meta: t.Sequence[t.Tuple[object, object]] = ()

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, object]) -> Task:
        outp = dict_.get("stdout_processor")
        errp = dict_.get("stderr_processor")
        pre = dict_.get("pre_run")
        post = dict_.get("post_run")
        if "command_factory" not in dict_:
            raise ValueError(f"invalid input data for {cls.__name__}")

        cf = locate(dict_["command_factory"])
        if not cf:
            raise ValueError(f"could not locate {repr(dict_['command_factory'])}")

        try:
            hosts = cls(
                **dict(
                    dict_,
                    command_factory=cf,
                    stdout_processor=locate(outp) if outp else None,
                    stderr_processor=locate(errp) if errp else None,
                    pre_run=locate(pre) if pre else None,
                    post_run=locate(post) if post else None,
                    meta=tuple(dict_.get("meta", {}).items()),
                )
            )
        except Exception:
            raise ValueError(f"invalid input data for {cls.__name__}")
        return hosts

    def to_dict(self) -> t.Dict[str, object]:
        cf = self.command_factory
        outp = self.stdout_processor
        errp = self.stderr_processor
        pre = self.pre_run
        post = self.post_run

        return dict(
            vars(self),
            command_factory=f"{cf.__module__}.{cf.__qualname__}",
            stdout_processor=f"{outp.__module__}.{outp.__qualname__}" if outp else None,
            stderr_processor=f"{errp.__module__}.{errp.__qualname__}" if errp else None,
            pre_run=f"{pre.__module__}.{pre.__qualname__}" if pre else None,
            post_run=f"{post.__module__}.{post.__qualname__}" if post else None,
            meta=dict(self.meta),
        )

    def results(self) -> Results:
        """Get the past results of this task.

        :rtype: Rviper.collections.esults
        """
        return self.pipe(Results.by_task)


@dataclass(frozen=True, order=True)
class Runner(Item):
    """A single task runner.

    A runner owns the responsibility of actually generating and running
    the command, store the result, calling, retrying, etc.
    We generally do not use it directly.
    """

    host: Host
    task: Task
    args: t.Sequence[str] = ()

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, object]) -> Runner:
        try:
            return cls(
                **dict(
                    dict_,
                    task=Task.from_dict(dict_["task"]),
                    host=Host.from_dict(dict_["host"]),
                    args=tuple(dict_.get("args", [])),
                )
            )
        except Exception:
            raise ValueError(f"invalid input data for {cls.__name__}")

    def to_dict(self) -> t.Dict[str, object]:
        return dict(vars(self), task=self.task.to_dict(), host=self.host.to_dict())

    def run(self, retry: int = 0, trigger_time: t.Optional[float] = None) -> Result:
        """Run the task on the host.

        :param int retry: Count of retries used.
        :param float trigger_time (optional): The trigger time used for grouping (auto generated).

        :rtype: viper.collections.Result
        """
        if not trigger_time:
            trigger_time = time()

        if not all(isinstance(a, str) for a in self.args):
            raise ValueError("{self.args}: args must be a list/tuple of strings.")

        command = self.task.command_factory(self.host, *self.args)

        if not command:
            raise ValueError(
                f"{self.task.command_factory} generated empty command ({command})."
            )

        if not all(isinstance(c, str) for c in command):
            raise ValueError(f"{command}: command must be a list/tuple of strings.")

        if self.task.pre_run:
            self.task.pre_run(self)

        start = time()

        try:
            r = subprocess.run(
                command,
                timeout=self.task.timeout,
                encoding="latin1",
                capture_output=True,
            )
            stdout, stderr, returncode = r.stdout, r.stderr, r.returncode
        except Exception as e:
            stdout, stderr, returncode = "", str(e), 123

        end = time()

        if self.task.stderr_processor:
            stdout = self.task.stdout_processor(stdout)

        if self.task.stderr_processor:
            stderr = self.task.stderr_processor(stderr)

        result = Result(
            trigger_time,
            self.task,
            self.host,
            self.args,
            command,
            stdout,
            stderr,
            returncode,
            start,
            end,
            retry,
        ).save()

        if self.task.post_run:
            self.task.post_run(result)

        if result.errored() and result.retry_left():
            return self.run(trigger_time=trigger_time, retry=retry + 1)

        return result

    def results(self) -> Results:
        """Fetch recent results of current runner from database.

        :rtype: viper.collections.Results
        """
        return self.pipe(Results.by_runner)


@dataclass(frozen=True)
class Runners(Items):
    _item_factory: t.Type[Runner] = field(init=False, default=Runner)

    def run(self, max_workers: int = Config.max_workers.value) -> Results:
        """Run the tasks.

        :param int max_workers: Maximum number of thread workers to use.

        :rtype: viper.collections.Results
        """

        trigger_time = time()

        if max_workers <= 1:
            # Run in sequence
            results = []
            for r in self.all():
                try:
                    results.append(r.run(trigger_time=trigger_time))
                except Exception:  # pragma: no cover
                    print(traceback.format_exc(), file=sys.stderr)
            return Results.from_items(*results)

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            # Run in parallel
            futures = [
                executor.submit(r.run, **{"trigger_time": trigger_time})
                for r in self._all
            ]
            results = []
            for f in as_completed(futures):
                try:
                    results.append(f.result())
                except Exception:  # pragma: no cover
                    print(traceback.format_exc(), file=sys.stderr)

        return Results.from_items(*results)

    def hosts(self) -> Hosts:
        """Get the list of hosts from the runners.

        :rtype: viper.collections.Hosts
        """
        return Hosts.from_items(t.host for t in self._all)


@dataclass(frozen=True, order=True)
class Result(Item):
    """The result of an executed task."""

    trigger_time: float
    task: Task
    host: Host
    args: t.Sequence[str]
    command: t.Sequence[str]
    stdout: str
    stderr: str
    returncode: int
    start: float
    end: float
    retry: int

    @classmethod
    def by_id(cls, id_: int) -> Result:
        """Fetch the result from DB by hash.

        :param int hash_: The hash value.

        :rtype: viper.collections.Result
        """

        with ViperDB(ViperDB.url) as conn:

            data = next(
                conn.execute(
                    """
                    SELECT
                        trigger_time, task, host, args, command, stdout, stderr,
                        returncode, start, end, retry
                    FROM results WHERE id = ?
                    """,
                    (id_,),
                )
            )

        return cls.from_dict(
            dict(
                trigger_time=data[0],
                task=loadjson(data[1]),
                host=loadjson(data[2]),
                args=tuple(loadjson(data[3])),
                command=tuple(loadjson(data[4])),
                stdout=data[5],
                stderr=data[6],
                returncode=data[7],
                start=data[8],
                end=data[9],
                retry=data[10],
            )
        )

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, object]) -> Task:
        try:
            return cls(
                **dict(
                    dict_,
                    args=tuple(dict_["args"]),
                    command=tuple(dict_["command"]),
                    task=Task.from_dict(dict_["task"]),
                    host=Host.from_dict(dict_["host"]),
                )
            )
        except Exception:
            raise ValueError(f"invalid input data for {cls.__name__}")

    def to_dict(self) -> t.Dict[str, object]:
        return dict(vars(self), task=self.task.to_dict(), host=self.host.to_dict())

    def ok(self) -> bool:
        """Returns True the result is success.

        :rtype: bool
        """
        return self.returncode == 0

    def errored(self) -> bool:
        """returns True if the result is failure.

        :rtype: bool
        """
        return self.returncode != 0

    def retry_left(self) -> int:
        """Get how many retries are left.

        :rtype: int
        """
        return self.task.retry - self.retry

    def save(self) -> Result:
        """Save the result in DB.

        :rtype: viper.collections.Result
        """

        with ViperDB() as conn:
            conn.execute(
                """
                INSERT INTO results (
                    hash, trigger_time, task, host, args, command,
                    stdout, stderr, returncode, start, end, retry
                ) VALUES (
                    ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?
                )
                """,
                (
                    self.hash(),
                    self.trigger_time,
                    self.task.to_json(),
                    self.host.to_json(),
                    dumpjson(self.args),
                    dumpjson(self.command),
                    self.stdout,
                    self.stderr,
                    self.returncode,
                    self.start,
                    self.end,
                    self.retry,
                ),
            )

        return self

    def runner(self) -> Runner:
        """Recreate the runner from the result.

        :rtype: viper.collections.Runner
        """
        return Runner(host=self.host, task=self.task, args=self.args)


@dataclass(frozen=True)
class Results(Items):
    """A group of :py:class:`viper.collections.Results`."""

    _item_factory: t.Type[Results] = field(init=False, default=Result)

    @classmethod
    def from_history(cls, final=False) -> Results:
        """Fetch and return all the results from history.

        :rtype: ciper.collections.Results
        """
        with ViperDB(ViperDB.url) as conn:
            rows = conn.execute("SELECT id FROM results ORDER BY start DESC")
            results = [cls._item_factory.by_id(r[0]) for r in rows]

        results = cls.from_items(*results)
        return results.final() if final else results

    @classmethod
    def by_host(cls, host: Host) -> Results:
        """Fetch and return results from history of the given host.

        :param viper.collections.Host host: Fetch results of this host.
        :rtype: viper.collections.Results
        """
        with ViperDB(ViperDB.url) as conn:
            rows = conn.execute(
                f"""
                SELECT id FROM results
                WHERE JSON_EXTRACT(host, '$.ip') = ?
                ORDER BY start DESC
                """,
                (host.ip,),
            )
            results = [cls._item_factory.by_id(r[0]) for r in rows]

        return cls.from_items(*results)

    @classmethod
    def by_task(cls, task: Task) -> Results:
        """Fetch and return results from history of the given task.

        :param viper.collections.Host host: Fetch results of this task.
        :rtype: viper.collections.Results
        """
        with ViperDB(ViperDB.url) as conn:
            rows = conn.execute(
                f"""
                SELECT id FROM results
                WHERE JSON_EXTRACT(task, '$.name') = ?
                ORDER BY start DESC
                """,
                (task.name,),
            )
            results = [cls._item_factory.by_id(r[0]) for r in rows]

        return cls.from_items(*results)

    def hosts(self) -> Hosts:
        """Get the list of hosts from the results.

        :rtype: viper.collections.Hosts
        """
        return Hosts.from_items(r.host for r in self._all)

    def final(self) -> Results:
        """Get the final results only (ignoring the previous retries).

        :rtype: viper.collections.Results
        """
        results: t.Dict[float, t.Dict[Host, Result]] = {}
        for result in self._all:
            if result.trigger_time not in results:
                results[result.trigger_time] = {result.host: result}
                continue

            res_grp = results[result.trigger_time]
            if result.host not in res_grp:
                res_grp[result.host] = result
                continue

            if res_grp[result.host].retry < result.retry:
                res_grp[result.host] = result

        out = []
        for res_grp in results.values():
            for res in res_grp.values():
                out.append(res)
        return Results.from_items(*out)

    def runners(self) -> Runners:
        """Recreate the runners from the results.

        :rtype: viper.collections.Runners
        """
        return Runners.from_items(r.runner() for r in self._all)

    def re_run(self, max_workers=Config.max_workers.value) -> Results:
        """Recreate the runners from the results and run again.

        :rtype: viper.collections.Results
        """
        return self.runners().run(max_workers=max_workers)



"""Common contants are defined here."""

from enum import Enum
from os import environ as env
from os import path

__all__ = ["Config"]


class Config(Enum):
    """Default viper configuration."""

    db_url = env.get("VIPER_DB_URL", "viperdb.sqlite3")
    max_workers = int(env.get("VIPER_MAX_WORKERS", 0))
    modules_path = path.expanduser(env.get("VIPER_MODULES_PATH", "."))



"""Viper Command-line Interface
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Demos and examples are probably the best way to explain viper CLI.

Let's get started.

Enable viper auto completion
----------------------------

.. code-block:: bash

    # Assuming you are using bash
    eval "$(viper autocomplete bash)"


Initialize current workspace (creates a `viperdb.sqlite3` file)
---------------------------------------------------------------

.. code-block:: bash

    viper init -f


Load hosts from file
--------------------

.. code-block:: bash

    viper hosts:from-file tests/data/hosts.csv --indent 4

    # or with a custom loader

    viper hosts:from-file tests/data/hosts.json --loader viper.demo.loaders.json --indent 4


Load hosts from a Python function
---------------------------------

.. code-block:: bash

    viper hosts viper.demo.hosts.group1 --indent 4


Count the number of hosts
-------------------------

.. code-block:: bash

    viper hosts viper.demo.hosts.group1 \\
            | viper hosts:count


Sort the hosts by custom logic
------------------------------

.. code-block:: bash

    viper hosts viper.demo.hosts.group1 \\
            | viper hosts:sort --key viper.demo.sort.by_ip -i 4


Pipe the hosts to a custom handler that formats the hosts to CSV
----------------------------------------------------------------

.. code-block:: bash

    viper hosts viper.demo.hosts.group1 \\
            | viper hosts:pipe viper.demo.handlers.hosts_to_csv


Let's save the hosts
--------------------

.. code-block:: bash

    viper hosts viper.demo.hosts.group1 > /tmp/hosts.json


Filter hosts
------------

.. code-block:: bash

    cat /tmp/hosts.json \\
            | viper hosts:filter viper.demo.filters.ip_is 8.8.8.8 --indent 4


Assign tasks to the given hosts
-------------------------------

.. code-block:: bash

    cat /tmp/hosts.json \\
            | viper hosts:task viper.demo.tasks.ping --indent 4


Run the assigned tasks
----------------------

.. code-block:: bash

    cat /tmp/hosts.json \\
            | viper hosts:task viper.demo.tasks.ping | viper runners:run --indent 4

    # or use a shortcut

    cat /tmp/hosts.json \\
            | viper hosts:run-task viper.demo.tasks.ping --indent 4


Run tasks in parallel using multiple workers
--------------------------------------------

.. code-block:: bash

    cat /tmp/hosts.json \\
            | viper hosts:run-task viper.demo.tasks.ping --max-workers 50 --indent 4


Get the past results from DB
----------------------------

.. code-block:: bash

    viper results


Get only the final results (ignoring the retries)
-------------------------------------------------

.. code-block:: bash

    viper results | viper results:final


Let's save the result
---------------------

.. code-block:: bash

    viper task viper.demo.tasks.ping \\
            | viper task:results > /tmp/results.json


Now filter the results by their status
--------------------------------------

Use the in-built ``where`` query

.. code-block:: bash

    # success
    cat /tmp/results.json \\
            | viper results:where returncode IS 0 -i 4

    # failed
    cat /tmp/results.json \\
            | viper results:where returncode IS_NOT 0 -i 4

Or use a custom filter

.. code-block:: bash

    # success
    cat /tmp/results.json \\
            | viper results:filter viper.demo.filters.result_ok -i 4

    # failed
    cat /tmp/results.json \\
            | viper results:filter viper.demo.filters.result_errored -i 4


Pipe the results to a custom handler
------------------------------------

.. code-block:: bash

    # print the status to terminal
    cat /tmp/results.json \\
            | viper results:pipe viper.demo.handlers.print_status

    # export the results to a csv file
    cat /tmp/results.json \\
            | viper results:pipe viper.demo.handlers.export_csv /tmp/results.csv


Let's do that again in one go
-----------------------------

.. code-block:: bash

    viper hosts viper.demo.hosts.group1 \\
            | viper hosts:run-task viper.demo.tasks.ping \\
            | viper results:pipe viper.demo.handlers.export_csv /tmp/results.csv


Get the unique trigger times from history (custom defined action)
-----------------------------------------------------------------

.. code-block:: bash

    viper lets viper.demo.actions.get_triggers
"""

from argparse import ArgumentParser
from argparse import Namespace
from pydoc import locate
from viper import __version__
from viper import Hosts
from viper import Results
from viper import Runners
from viper import Task
from viper.cli_base import SubCommand
from viper.collections import Collection as ViperCollection
from viper.collections import FilterType
from viper.collections import HandlerType
from viper.collections import WhereConditions
from viper.const import Config
from viper.db import ViperDB

import os
import sqlite3
import sys
import traceback
import typing as t

__all__ = ["func", "run"]


def func(objpath: str) -> t.Union[HandlerType, FilterType]:
    """Resolved python function from given string."""

    funcobj = locate(objpath)
    if not func:
        raise ValueError(f"could not resolve {repr(objpath)}.")

    if not callable(funcobj):
        raise ValueError(f"{repr(objpath)} is not a valid function.")

    return funcobj


class AutocompleteCommand(SubCommand):
    """generate the auto completion script"""

    name = "autocomplete"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "shell", help="the target shell", choices=["bash", "tcsh", "zsh", "fish"]
        )

    def __call__(self, args: Namespace) -> int:
        import subprocess

        try:
            import argcomplete  # noqa F401
        except ImportError:
            raise RuntimeError(
                '"argcomplete" is missing! run `pip install -U "viper-infra-commander[autocomplete]"`'
            )

        _shell = "bash" if args.shell == "zsh" else args.shell
        script = subprocess.run(
            ["register-python-argcomplete", "viper", "-s", _shell],
            stdout=subprocess.PIPE,
            check=True,
        ).stdout.decode("latin1")

        if args.shell == "zsh":
            script = f"autoload bashcompinit\nbashcompinit\n{script}"

        print(script)
        return 0


class InitCommand(SubCommand):
    """initialize the current workspace"""

    name = "init"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="remove or overwrite existing data",
        )

    def __call__(self, args: Namespace) -> int:
        try:
            ViperDB.init(ViperDB.url, force=args.force)
        except sqlite3.OperationalError:
            raise RuntimeError(
                "database already exists!"
                " use '-f'/'--force' to force re-create the database."
            )
        return 0


class LetsCommand(SubCommand):
    """perform any defined action"""

    name = "lets"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("action", type=func, help="action definition location")
        parser.add_argument("args", nargs="*", help="optional arguments for the action")

    def __call__(self, args: Namespace) -> int:
        res = args.action(*args.args)
        if res is not None:
            print("\n".join(map(str, res)) if isinstance(res, tuple) else res)
        return 0


class RunJobCommand(SubCommand):
    """[Hosts -> Results] run a job on the given hosts"""

    name = "run-job"
    aliases = ("run",)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("job", type=func, help="job definition location")
        parser.add_argument("args", nargs="*", help="arguments to be passed to the job")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        results = args.job(Hosts.from_json(input()), *args.args)
        if not isinstance(results, Results):
            raise ValueError(
                f"a job must return {Results} object but got {type(results)}"
            )
        print(results.to_json(indent=args.indent))
        return 0


class TaskFromFuncCommand(SubCommand):
    """[-> Task] get the task from a Python function location"""

    name = "task:from-func"
    aliases = ("task",)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "func", type=Task.from_func, help=Task.from_func.__doc__.lower()
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(args.func.to_json(indent=args.indent))
        return 0


class TaskResultsCommand(SubCommand):
    """[Task -> Results] get the past results of given task"""

    name = "task:results"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Task.from_json(input()).results().to_json(indent=args.indent))
        return 0


class TaskFormatCommand(SubCommand):
    """[Task -> str] format the data using the given template"""

    name = "task:format"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("template", help="use Python's string format syntax")

    def __call__(self, args: Namespace) -> int:
        print(Task.from_json(input()).format(args.template))
        return 0


class HostsFromFuncCommand(SubCommand):
    """[-> Hosts] get a group of hosts from a Python function location"""

    name = "hosts:from-func"
    aliases = ("hosts",)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "func", type=Hosts.from_func, help=Hosts.from_func.__doc__.lower()
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(args.func.to_json(indent=args.indent))
        return 0


class HostsFromFileCommand(SubCommand):
    """[-> Hosts] get a group of hosts from a file"""

    name = "hosts:from-file"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filepath")
        parser.add_argument(
            "--loader",
            type=func,
            help="function that resolves a file object to a viper.Hosts object",
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Hosts.from_file(args.filepath, loader=args.loader).to_json(
                indent=args.indent
            )
        )
        return 0


class HostsTaskCommand(SubCommand):
    """[Hosts -> Runners] assign a task to each host"""

    name = "hosts:task"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "task", type=Task.from_func, help=Task.from_func.__doc__.lower()
        )
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the command factory"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:

        print(
            Hosts.from_json(input())
            .task(args.task, *args.args)
            .to_json(indent=args.indent)
        )
        return 0


class HostsRunTaskCommand(SubCommand):
    """[Hosts -> Results] assign a task to each host and run"""

    name = "hosts:run-task"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "task", type=Task.from_func, help=Task.from_func.__doc__.lower()
        )
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the command factory"
        )
        parser.add_argument("--max-workers", type=int, default=Config.max_workers.value)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Hosts.from_json(input())
            .run_task(args.task, *args.args, max_workers=args.max_workers)
            .to_json(indent=args.indent)
        )
        return 0


class HostsFilterCommand(SubCommand):
    """[Hosts -> Hosts] filter hosts by a given function"""

    name = "hosts:filter"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filter", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Hosts.from_json(input())
            .filter(args.filter, *args.args)
            .to_json(indent=args.indent)
        )
        return 0


class HostsCountCommand(SubCommand):
    """[Hosts -> int] count the number of hosts"""

    name = "hosts:count"

    def add_arguments(self, parser: ArgumentParser) -> None:
        pass

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_json(input()).count())
        return 0


class HostsSortCommand(SubCommand):
    """[Hosts -> Hosts] sort the hosts"""

    name = "hosts:sort"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("--key", type=func)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_json(input()).sort(key=args.key).to_json(indent=args.indent))
        return 0


class HostsPipeCommand(SubCommand):
    """[Hosts -> ?] pipe the hosts to the given handler"""

    name = "hosts:pipe"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("handler", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        obj = Hosts.from_json(input()).pipe(args.handler, *args.args)

        if obj is None:
            return 0

        if isinstance(obj, ViperCollection):
            print(obj.to_json(indent=args.indent))
            return 0

        print(obj)
        return 0


class HostsFormatCommand(SubCommand):
    """[Hosts -> str] format the data using the given template"""

    name = "hosts:format"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("template", help="use Python's string format syntax")
        parser.add_argument(
            "-s", "--sep", help="separator used to join the strings", default="\n"
        )

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_json(input()).format(args.template, sep=args.sep))
        return 0


class HostsWhereCommand(SubCommand):
    """[Hosts -> Hosts] select hosts matching the given query"""

    name = "hosts:where"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("key", help="property name/path of each item")
        parser.add_argument(
            "condition", choices=[o.value for o in WhereConditions],
        )
        parser.add_argument("values", nargs="*")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Hosts.from_json(input())
            .where(args.key, WhereConditions(args.condition), args.values)
            .to_json(indent=args.indent)
        )
        return 0


class HostsHeadCommand(SubCommand):
    """[Hosts -> Hosts] get the first 'n' hosts"""

    name = "hosts:head"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_json(input()).head(args.n).to_json(indent=args.indent))
        return 0


class HostsTailCommand(SubCommand):
    """[Hosts -> Hosts] get the last 'n' hosts"""

    name = "hosts:tail"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_json(input()).tail(args.n).to_json(indent=args.indent))
        return 0


class HostsResultsCommand(SubCommand):
    """[Hosts -> Results] get the past results of the hosts"""

    name = "hosts:results"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_json(input()).results().to_json(indent=args.indent))
        return 0


class RunnersFilterCommand(SubCommand):
    """[Runners -> Runners] filter runners by a given function"""

    name = "runners:filter"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filter", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Runners.from_json(input())
            .filter(args.filter, *args.args)
            .to_json(indent=args.indent)
        )
        return 0


class RunnersCountCommand(SubCommand):
    """[Runners -> int] count the number of runners"""

    name = "runners:count"

    def add_arguments(self, parser: ArgumentParser) -> None:
        pass

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_json(input()).count())
        return 0


class RunnersSortCommand(SubCommand):
    """[Runners -> Runners] sort the runners"""

    name = "runners:sort"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("--key", type=func)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_json(input()).sort(key=args.key).to_json(indent=args.indent))
        return 0


class RunnersPipeCommand(SubCommand):
    """[Runners -> ?] pipe the runners to the given handler"""

    name = "runners:pipe"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("handler", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        obj = Runners.from_json(input()).pipe(args.handler, *args.args)

        if obj is None:
            return 0

        if isinstance(obj, ViperCollection):
            print(obj.to_json(indent=args.indent))
            return 0

        print(obj)
        return 0


class RunnersFormatCommand(SubCommand):
    """[Runners -> str] format the data using the given template"""

    name = "runners:format"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("template", help="use Python's string format syntax")
        parser.add_argument(
            "-s", "--sep", help="separator used to join the strings", default="\n"
        )

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_json(input()).format(args.template, sep=args.sep))


class RunnersWhereCommand(SubCommand):
    """[Runners -> Runners] select runners matching the given query"""

    name = "runners:where"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("key", help="property name/path of each item")
        parser.add_argument(
            "condition", choices=[o.value for o in WhereConditions],
        )
        parser.add_argument("values", nargs="*")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Runners.from_json(input())
            .where(args.key, WhereConditions(args.condition), args.values)
            .to_json(indent=args.indent)
        )
        return 0


class RunnersHeadCommand(SubCommand):
    """[Runners -> Runners] get the first 'n' runners"""

    name = "runners:head"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_json(input()).head(args.n).to_json(indent=args.indent))
        return 0


class RunnersTailCommand(SubCommand):
    """[Runners -> Runners] get the last 'n' runners"""

    name = "runners:tail"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_json(input()).tail(args.n).to_json(indent=args.indent))
        return 0


class RunnersRunCommand(SubCommand):
    """[Runners -> Results] run the assigned tasks"""

    name = "runners:run"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the command factory"
        )
        parser.add_argument("--max-workers", type=int, default=Config.max_workers.value)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Runners.from_json(input())
            .run(*args.args, max_workers=args.max_workers)
            .to_json(indent=args.indent)
        )
        return 0


class RunnersHostsCommand(SubCommand):
    """[Runners -> Hosts] get the hosts from the runners"""

    name = "runners:hosts"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_json(input()).hosts().to_json(indent=args.indent))
        return 0


class ResultsFromHistoryCommand(SubCommand):
    """[-> Results] get the past results from database"""

    name = "results:from-history"
    aliases = ("results",)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--final",
            action="store_true",
            help="get the final results only (shortcut to `viper results | viper results:final`)",
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_history(final=args.final).to_json(indent=args.indent))
        return 0


class ResultsFilterCommand(SubCommand):
    """[Results -> Results] filter results by a given handler"""

    name = "results:filter"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filter", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Results.from_json(input())
            .filter(args.filter, *args.args)
            .to_json(indent=args.indent)
        )
        return 0


class ResultsCountCommand(SubCommand):
    """[Results -> int] count the number of results"""

    name = "results:count"

    def add_arguments(self, parser: ArgumentParser) -> None:
        pass

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).count())
        return 0


class ResultsSortCommand(SubCommand):
    """[Results -> Results] sort the results"""

    name = "results:sort"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("--key", type=func)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).sort(key=args.key).to_json(indent=args.indent))
        return 0


class ResultsPipeCommand(SubCommand):
    """[Results -> ?] pipe the results to the given handler"""

    name = "results:pipe"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("handler", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        obj = Results.from_json(input()).pipe(args.handler, *args.args)

        if obj is None:
            return 0

        if isinstance(obj, ViperCollection):
            print(obj.to_json(indent=args.indent))
            return 0

        print(obj)
        return 0


class ResultsFormatCommand(SubCommand):
    """[Results -> str] format the data using the given template"""

    name = "results:format"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("template", help="use Python's string format syntax")
        parser.add_argument(
            "-s", "--sep", help="separator used to join the strings", default="\n"
        )

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).format(args.template, sep=args.sep))
        return 0


class ResultsWhereCommand(SubCommand):
    """[Results -> Results] select results matching the given query"""

    name = "results:where"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("key", help="property name of each item")
        parser.add_argument(
            "condition",
            choices=[o.value for o in WhereConditions],
            help="condition to apply",
        )
        parser.add_argument("values", nargs="*", help="values for the key")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Results.from_json(input())
            .where(args.key, WhereConditions(args.condition), args.values)
            .to_json(indent=args.indent)
        )
        return 0


class ResultsHeadCommand(SubCommand):
    """[Results -> Results] get the first 'n' results"""

    name = "results:head"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).head(args.n).to_json(indent=args.indent))
        return 0


class ResultsTailCommand(SubCommand):
    """[Results -> Results] get the last 'n' results"""

    name = "results:tail"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).tail(args.n).to_json(indent=args.indent))
        return 0


class ResultsHostsCommand(SubCommand):
    """[Results -> Hosts] get the hosts from the results"""

    name = "results:hosts"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).hosts().to_json(indent=args.indent))
        return 0


class ResultsRunnersCommand(SubCommand):
    """[Results -> Runners] recreate the runners from the results"""

    name = "results:runners"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).runners().to_json(indent=args.indent))
        return 0


class ResultsReRunCommand(SubCommand):
    """[Results -> Results] recreate the runners from the results and run again"""

    name = "results:re-run"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("--max-workers", type=int, default=Config.max_workers.value)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Results.from_json(input())
            .runners()
            .run(max_workers=args.max_workers)
            .to_json(indent=args.indent)
        )
        return 0


class ResultsByTaskCommand(SubCommand):
    """[Task -> Results] get the past results of given task"""

    name = "results:by-task"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.by_task(Task.from_json(input())).to_json(indent=args.indent))
        return 0


class ResultsFinal(SubCommand):
    """[Results -> Results] get the final results only (ignoring the previous retries)"""

    name = "results:final"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).final().to_json(indent=args.indent))
        return 0


def run() -> int:
    parser = ArgumentParser(
        "viper", description=f"Viper Infrastructure Commander {__version__}"
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--debug",
        action="store_true",
        help="show traceback information when an exception is raised",
    )
    subparsers = parser.add_subparsers()

    # Autocomplete command
    AutocompleteCommand.attach_to(subparsers)

    # Init command
    InitCommand.attach_to(subparsers)

    # Init command
    LetsCommand.attach_to(subparsers)

    # Job run command
    RunJobCommand.attach_to(subparsers)

    # Task commands
    TaskFromFuncCommand.attach_to(subparsers)
    TaskResultsCommand.attach_to(subparsers)
    TaskFormatCommand.attach_to(subparsers)

    # Hosts commands
    HostsFromFileCommand.attach_to(subparsers)
    HostsFromFuncCommand.attach_to(subparsers)

    HostsFilterCommand.attach_to(subparsers)
    HostsCountCommand.attach_to(subparsers)
    HostsSortCommand.attach_to(subparsers)
    HostsPipeCommand.attach_to(subparsers)
    HostsFormatCommand.attach_to(subparsers)
    HostsWhereCommand.attach_to(subparsers)
    HostsHeadCommand.attach_to(subparsers)
    HostsTailCommand.attach_to(subparsers)

    HostsTaskCommand.attach_to(subparsers)
    HostsRunTaskCommand.attach_to(subparsers)
    HostsResultsCommand.attach_to(subparsers)

    # Task runners commands
    RunnersFilterCommand.attach_to(subparsers)
    RunnersCountCommand.attach_to(subparsers)
    RunnersSortCommand.attach_to(subparsers)
    RunnersPipeCommand.attach_to(subparsers)
    RunnersFormatCommand.attach_to(subparsers)
    RunnersWhereCommand.attach_to(subparsers)
    RunnersHeadCommand.attach_to(subparsers)
    RunnersTailCommand.attach_to(subparsers)

    RunnersRunCommand.attach_to(subparsers)
    RunnersHostsCommand.attach_to(subparsers)

    # Task results commands
    ResultsFromHistoryCommand.attach_to(subparsers)

    ResultsFilterCommand.attach_to(subparsers)
    ResultsCountCommand.attach_to(subparsers)
    ResultsSortCommand.attach_to(subparsers)
    ResultsPipeCommand.attach_to(subparsers)
    ResultsFormatCommand.attach_to(subparsers)
    ResultsWhereCommand.attach_to(subparsers)
    ResultsHeadCommand.attach_to(subparsers)
    ResultsTailCommand.attach_to(subparsers)

    ResultsFinal.attach_to(subparsers)
    ResultsHostsCommand.attach_to(subparsers)
    ResultsRunnersCommand.attach_to(subparsers)
    ResultsReRunCommand.attach_to(subparsers)
    ResultsByTaskCommand.attach_to(subparsers)

    # Add the modules path
    sys.path.insert(0, os.path.realpath(Config.modules_path.value))

    # Import project specific commands
    if os.path.exists("viperfile.py"):
        from viper import project

        sys.path.insert(0, os.path.realpath("."))
        import viperfile

        for k, v in vars(viperfile).items():
            if not isinstance(v, project.Project):
                continue

            for subcommand in v.all_commands():
                subcommand.attach_to(subparsers)

    try:
        # Add support for auto completion
        import argcomplete

        argcomplete.autocomplete(parser)
    except ImportError:
        pass

    args = parser.parse_args()

    if not hasattr(args, "_handler"):
        parser.print_usage()
        return 1

    try:
        return args._handler(args)
    except Exception as e:
        print(f"error: {e}", file=sys.stderr)
        if args.debug:
            print(traceback.format_exc(), file=sys.stderr)
        return 1



from viper import Result
from viper import Runner

import sys


def log_command(runner: Runner) -> None:
    """Log the command before run."""
    command = runner.task.command_factory(runner.host, *runner.args)
    print(runner.host.ip, command, file=sys.stderr, sep=": ")


def log_status(result: Result) -> None:
    """Log the status after task run."""

    if result.ok():
        print(result.host.ip, "OK", file=sys.stderr, sep=": ")
        return

    print(result.host.ip, "FAILED", file=sys.stderr, sep=": ")



from viper import Host
from viper import Hosts
from viper import Result
from viper import Results
from viper import Runners

import typing as t


def ip_is(host: Host, ip: str) -> bool:
    """Filter a host by IP address."""
    return host.ip == ip


def result_ok(result: Result) -> bool:
    """Filter results by OK status."""
    return result.ok()


def result_errored(result: Result) -> bool:
    """Filter results by error status."""
    return result.errored()


def exact(obj: t.Union[Hosts, Runners, Results], key: str, val: str) -> bool:
    """A generic filter for all items."""
    return hasattr(obj, key) and str(getattr(obj, key)) == val


def contains(obj: t.Union[Hosts, Runners, Results], key: str, val: str) -> bool:
    """A generic filter for all items."""
    return hasattr(obj, key) and val in str(getattr(obj, key))


def startswith(obj: t.Union[Hosts, Runners, Results], key: str, val: str) -> bool:
    """A generic filter for all items."""
    return hasattr(obj, key) and str(getattr(obj, key)).startswith(val)


def endswith(obj: t.Union[Hosts, Runners, Results], key: str, val: str) -> bool:
    """A generic filter for all items."""
    return hasattr(obj, key) and str(getattr(obj, key)).endsswith(val)



# -*- coding: utf-8 -*-
from os import path
from setuptools import find_packages
from setuptools import setup
from viper import __author__
from viper import __description__
from viper import __email__
from viper import __homepage__
from viper import __license__
from viper import __version__

import typing as t

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

install_requires: t.List[str] = []
autocomplete_requires = ["argcomplete==1.10.3"]
testing_requires = autocomplete_requires + [
    "pytest>=4.4.1",
    "pytest-cov>=2.7.1",
    "black>=19.3b0",
    "isort>=4.3.21",
    "mypy>=0.710",
    "lxml>=4.3.4",
    "pre-commit>=1.20.0",
    "pre-commit-hooks>=2.4.0",
]
dev_requires = testing_requires + ["tox>=3.12.1", "twine>=3.1.1", "sphinx>=2.2.1"]

setup(
    name="viper-infra-commander",
    version=__version__,
    description=__description__,
    long_description=long_description,
    url=__homepage__,
    author=__author__,
    author_email=__email__,
    license=__license__,
    py_modules=["viper"],
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators ",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    platforms=["Any"],
    zip_safe=False,
    keywords="viper infrastructure commander",
    packages=find_packages(exclude=["contrib", "docs", "tests", "examples"]),
    install_requires=install_requires,
    extras_require={
        "autocomplete": autocomplete_requires,
        "testing": testing_requires,
        "dev": dev_requires,
    },
    entry_points={"console_scripts": ["viper = viper.main:main"]},
)



"""Viper Project APIs (the ``viperfile.py``)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Viper provides APIs for customizing the viper CLI based on the project
requirement. When viper detects a ``viperfile.py`` file in the current
working directory, it scans that file and adds the project specific
commands and options to it's own list.

We can see the commands by running ``viper --help``. These commands are
prefixed with ``@`` so that they are easily recognizable.


After defining this in ``viperfile.py`` we can use commands like ``viper @myproj:allhosts``
(example) to get the list of hosts which can be piped to other commands that recieves
a list of hosts from `stdin`.


.. tip:: See :py:mod:`viper.demo.viperfile` for the full project example.
"""


from __future__ import annotations
from argparse import ArgumentParser
from argparse import Namespace
from dataclasses import dataclass
from dataclasses import field
from viper.cli_base import SubCommand
from viper.collections import Collection as ViperCollection
from viper.collections import Hosts
from viper.collections import Items
from viper.collections import Results

import typing as t

ArgType = t.Tuple[t.Sequence[str], t.Dict[str, object]]


def arg(*args: str, **kwargs: object) -> ArgType:
    """Argumenst to be passed to argparse"""
    return args, kwargs


@dataclass
class Project:
    """A project (namespace) for the viper sub commands

    :param str prefix: Prefix for the related sub commands.

    When we define a project, we basically define a namespace (a prefix)
    for the commands.

    .. tip:: See :py:mod:`viper.demo.viperfile`
    """

    prefix: str
    action_commands: t.List[SubCommand] = field(default_factory=lambda: [])
    hostgroup_commands: t.List[SubCommand] = field(default_factory=lambda: [])
    filter_commands: t.List[SubCommand] = field(default_factory=lambda: [])
    handler_commands: t.List[SubCommand] = field(default_factory=lambda: [])
    job_commands: t.List[SubCommand] = field(default_factory=lambda: [])

    def all_commands(self) -> t.List[SubCommand]:
        """Return all sub commands"""
        return (
            self.action_commands
            + self.hostgroup_commands
            + self.filter_commands
            + self.handler_commands
            + self.job_commands
        )

    def hostgroup(self, args: t.Optional[t.Sequence[ArgType]] = None) -> Hosts:
        """Use this decorator to define host groups

        :param list args (optional): Arguments to be parsed by py:class:`argparse.ArgumentParser`

        .. tip:: See :py:func:`viper.demo.viperfile.allhosts`.
        """

        def wrapper(
            func: t.Callable[[Namespace], Hosts]
        ) -> t.Callable[[Namespace], Hosts]:

            doc = func.__doc__.splitlines()[0] if func.__doc__ else ""

            class HostGroupCommand(SubCommand):
                __doc__ = f"[-> Hosts] {doc}"
                name = f"@{self.prefix}:{func.__name__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg[0], **arg[1])
                    parser.add_argument("-i", "--indent", type=int, default=None)

                def __call__(self, args: Namespace) -> int:
                    print(func(args).to_json(indent=args.indent))
                    return 0

            self.hostgroup_commands.append(HostGroupCommand)
            return func

        return wrapper

    def filter(
        self, objtype: t.Type[Items], args: t.Optional[t.Sequence[ArgType]] = None
    ):
        """Use this decorator to define filters.

        :param type objtype: The object type that the filter is expecting.
        :param list args (optional): List of arguments for :py:class:`argparse.ArgumentParser`.

        .. tip:: See :py:func:`viper.demo.viperfile.hosts_by`.
        """

        if not issubclass(objtype, Items):
            raise ValueError(f"{objtype} does not have filter option")

        def wrapper(
            func: t.Callable[[Items, Namespace], bool],
        ) -> t.Callable[[Items, Namespace], bool]:

            doc = func.__doc__.splitlines()[0] if func.__doc__ else ""

            class FilterCommand(SubCommand):
                __doc__ = f"[{objtype.__name__} -> {objtype.__name__}] {doc}"
                name = f"@{self.prefix}:{func.__name__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg[0], **arg[1])
                    parser.add_argument("-i", "--indent", type=int, default=None)

                def __call__(self, args: Namespace) -> int:
                    print(
                        objtype.from_json(input())
                        .filter(lambda host: func(host, args))
                        .to_json(indent=args.indent)
                    )
                    return 0

            self.filter_commands.append(FilterCommand)
            return func

        return wrapper

    def handler(
        self,
        fromtype: ViperCollection,
        totype: t.Optional[type] = None,
        args: t.Optional[t.Sequence[ArgType]] = None,
    ):
        """Use this decorator to define handlers

        :param type fromtype: The type of object this handler is expecting.
        :param type totype: The type of object this handler returns.
        :param list args (optional): List of arguments for :py:class:`argparse.ArgumentParser`.

        .. tip:: See :py:func:`viper.demo.viperfile.results2csv`.
        """

        if not issubclass(fromtype, ViperCollection):
            raise ValueError(f"{fromtype} does not have pipe option")

        def wrapper(
            func: t.Callable[[ViperCollection, Namespace], object]
        ) -> t.Callable[[ViperCollection, Namespace], object]:

            doc = func.__doc__.splitlines()[0] if func.__doc__ else ""

            class HandlerCommand(SubCommand):
                __doc__ = f"[{fromtype.__name__} -> {totype.__name__}] {doc}"
                name = f"@{self.prefix}:{func.__name__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg[0], **arg[1])
                    parser.add_argument("-i", "--indent", type=int, default=None)

                def __call__(self, args: Namespace) -> int:
                    obj = fromtype.from_json(input()).pipe(lambda obj: func(obj, args))
                    if totype:
                        print(
                            obj.to_json(indent=args.indent)
                            if isinstance(obj, ViperCollection)
                            else totype(obj)
                        )
                    return 0

            self.handler_commands.append(HandlerCommand)
            return func

        return wrapper

    def job(
        self, args: t.Optional[t.Sequence[ArgType]] = None,
    ):
        """Use this decorator to define a job.

        :param list args (optional): List of arguments for :py:class:`argparse.ArgumentParser`.

        .. tip:: See :py:func:`viper.demo.viperfile.remote_exec`.
        """

        def wrapper(
            func: t.Callable[[ViperCollection, Namespace], Results]
        ) -> t.Callable[[ViperCollection, Namespace], Results]:

            doc = func.__doc__.splitlines()[0] if func.__doc__ else ""

            class JobCommand(SubCommand):
                __doc__ = f"[Hosts -> Results] {doc}"
                name = f"@{self.prefix}:{func.__name__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg[0], **arg[1])
                    parser.add_argument("-i", "--indent", type=int, default=None)

                def __call__(self, args: Namespace) -> int:
                    results = func(Hosts.from_json(input()), args)
                    if not isinstance(results, Results):
                        raise ValueError(
                            f"a job must return {Results} object but got {type(results)}"
                        )
                    print(results.to_json(indent=args.indent))
                    return 0

            self.job_commands.append(JobCommand)
            return func

        return wrapper

    def action(
        self, args: t.Optional[t.Sequence[ArgType]] = None,
    ):
        """Use this decorator to define an action.

        :param list args (optional): List of arguments for :py:class:`argparse.ArgumentParser`.

        .. tip:: See :py:func:`viper.demo.viperfile.get_triggers`.
        """

        def wrapper(
            func: t.Callable[[Namespace], object]
        ) -> t.Callable[[Namespace], object]:

            doc = func.__doc__.splitlines()[0] if func.__doc__ else ""

            class ActionCommand(SubCommand):
                __doc__ = f"{doc}"
                name = f"@{self.prefix}:{func.__name__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg[0], **arg[1])

                def __call__(self, args: Namespace) -> int:
                    res = func(args)
                    if res is not None:
                        print(
                            "\n".join(map(str, res)) if isinstance(res, tuple) else res
                        )
                    return 0

            self.action_commands.append(ActionCommand)
            return func

        return wrapper
